
## License ##
#######################################################################
# Copyright (c) 2020 Dr. Dr. Christoph Freier, & Ian Taylor BD Life 
# Sciences - Informatics, Ashland, OR
#
# R Script Template Designed and Written by: Josh Luthy and Ian Taylor
#
# License
# The software is distributed under the terms of the 
# Artistic License 2.0
# http://www.r-project.org/Licenses/Artistic-2.0
# 
# Disclaimer
# This software and documentation come with no warranties of any kind.
# This software is provided "as is" and any express or implied 
# warranties, including, but not limited to, the implied warranties of
# merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the  copyright holder be liable for any direct, 
# indirect, incidental, special, exemplary, or consequential damages
# (including but not limited to, procurement of substitute goods or 
# services; loss of use, data or profits; or business interruption)
# however caused and on any theory of liability, whether in contract,
# strict liability, or tort arising in any way out of the use of this 
# software.
######################################################################







####
## Functions
####

packages<-function(x){
  x<-as.character(match.call()[[2]])
  if(!require(x,character.only=TRUE)){
    install.packages(pkgs=x,repos="http://cran.r-project.org")
    require(x,character.only=TRUE)
  }
}

devPackages<-function(x){
  if (!require(x,character.only=TRUE)){
    suppressMessages(suppressWarnings(devtools::install_github(x)))
  }else{
    require(x,character.only=TRUE)
  }
}

fCorePack<-function(x){
  x<-as.character(match.call()[[2]])
  if (!require(x,character.only=TRUE)){
    BiocManager::install(x)
    require(x,character.only=TRUE)
  }
}

stopAndReturnError <- function(message, ...) {
  cat(message, file=paste0(outs, "/Error_Message_from_R.txt"))
  stop(message, ...)
}

dip_it <- function(data, ...) {
  # if (!class(data) == "numeric")
  #   stop("data should be a numeric!")
  if (length(data) <= 1)
    stop("data length should be >1!")
  if (is.unsorted(data))
    data <- sort(data)
  n <- as.integer(length(data))
  result <- .C("diptst",
               x	      = as.double(data),
               n	      = n,
               dip      = double(1),
               lo.hi    = integer(4),
               ifault   = integer(1),
               gcm      = integer(n),
               lcm      = integer(n),
               mn	      = integer(n),
               mj	      = integer(n),
               debug    = as.integer(FALSE),
               min.is.0 = as.logical(FALSE))[3]
  return(result[[1]])
}

check_modality_it <- function(data,
                              cluster = "cluster", ...) {
  if(!is.data.frame(data))
    stop("Wrong input data, should be a data frame!")
  if(!cluster %in% colnames(data))
    stop("Missing input data, need to know cluster!")
  result <- data %>%
    rename(., cluster = cluster) %>%
    group_by(cluster) %>%
    summarise_if(is.numeric, list(~dip_it(.)))
  return(result)
}

heatmap_visual_it <- function(data,
                           cluster = "cluster", ...) {
  if(!is.data.frame(data))
    stop("Wrong input data, should be a data frame!")
  if(!cluster %in% colnames(data))
    stop("Missing input data, need to know cluster!")
  result <- data %>%
    rename(., cluster = cluster) %>%
    group_by(cluster) %>%
    pivot_longer(-cluster, names_to = "protein", values_to = "test") %>%
    ggplot(aes(x = protein, y = as.factor(cluster), fill = test)) +
  ggplot2::geom_tile(color = "white", size = .5) +
    scale_fill_gradient(low = "white", high = "red", limits=c(0, (max(data))/20)) +
    theme_minimal() +
    coord_flip() +
    theme(legend.title = element_blank(),
          axis.title   = element_blank())
  return(result)
}


arrange_col <- function(dl)
{
  n_col <- ncol(dl[[1]])
  out_names <- names(dl[[1]])
  out <- lapply(1:n_col, function(i, x) as.data.frame(lapply(x, '[[' , i = i)) , dl)
  names(out) <- out_names
  return(out)
}


#####
## Install
#####

packages(devtools)
devPackages("Kawameicha/clustRcheck")
packages(tidyverse)
packages(clustRcheck)
packages(dplyr)
packages(tidyr)
packages(ggplot2)
packages(viridis)
packages(magma)
packages(plasma)

# install.packages("devtools")
# devtools::install_github("Kawameicha/clustRcheck")
# library(tidyverse)
# library(clustRcheck)

#####
## Import
#####

outPath <- "/Users/10201397/Documents/Code/euclid/demoData"
inCSVPath <- "/Users/10201397/Documents/Code/euclid/demoData/Abseq_Demo.csv.Cell_Type.ExtNode.csv"
catParRaw <- "[Cell_Type]"
runClustRCheckBool <- TRUE
runResolutionIndexBool <- TRUE

#####
#####



#####
## Massage Imports
#####

noKatPar <- TRUE
if(catParRaw != "[]" && catParRaw != ""){
  catParam <- gsub("(\\[)","",gsub("(\\])","",catParRaw))
  catPar <- as.symbol(catParam)
  noKatPar <- FALSE
}

input00 <- read.csv(inCSVPath, sep = ",", header = TRUE, check.names = FALSE)

if(!nrow(input00) > 0){
  print("Unable to read the input file correctly, try again.")
  stopAndReturnError("Unable to read the input file correctly.", call. = FALSE)
}
if ("EventNumberDP" %in% colnames(input00)){
  print("EventNumber Detected - This is FlowJo")
  input00 <- input00 %>%
    select(EventNumberDP, everything())
  colnames(input00) <- gsub( "(^.*)( :: )" , "", colnames(input00), ignore.case = TRUE )
  isSeqGeqBool <- F
} else {
  print("We're in SeqGeq")
  isSeqGeqBool <- T
}

if (isSeqGeqBool == F) {
  input00$EventNumberDP <- NULL
}

## Check if the user selected any parameters (todo - unless no parameters is a valid choice in some case? - todo)
if (!length(input00) > 0) {
  stopAndReturnError("Forget something?...\nNo parameters for plotting detected in R.\n\nTry deleting and then 
       restarting the plugin,\nmaking sure to select some parameters in\nthe top parameter selector dialog there.\n\nCheers!")
}
numUniqueParams <- length(unique(col(input00)))


#####
## Run
#####

if(runClustRCheckBool){   # Impliment functions for ClustRCheck
  d <- check_modality_it(input00, cluster = catParam)
  p <- heatmap_visual_it(input00, cluster = catParam)
  ggsave(paste0(outPath, "/ClustRCheck_HeatMap.png"),  plot = p, dpi = 320, device = "png")
}
    

if(runResolutionIndexBool){   # Impliment functions for ResolutionIndex

  # Lets take the selected Modal Parameter column and move it to the front of the DF before running gather()
  input0 <- input00 %>% dplyr::select(catParam, everything())
  
  # Get rid of pesky clusters at Zero (those should be cells not included in clustering which snuck in here)
  input <- subset(input0, input0[[catPar]]!=0)
  
  
  # Use gather function to quickly create the data frame. (I don't think we actually use this currently)
  # DataFrame <- gather(input, key = "Param", value = "Data", 2:ncol(input))
  
  
  cats <- unique(input[catParam])
  
  colNamer <- colnames(input)
  
  m <- length(colNamer)
  
  centroids <- input %>% 
    group_by_at(vars((!!catPar))) %>%
    summarise_all(mean)
  
  clusterWidth <- input %>% 
    group_by_at(vars((!!catPar))) %>% 
    summarise_all( sd ) %>%
    mutate(clusterWidth = sqrt(rowSums(.[2:m])^2))
  
  centroidMean <- as.data.frame(centroids)
  
  # n choose k: n!/[(n-k)!k!]
  # our k is always 2 (pairs)
  # this will be our new number of rows when comparing each possible pair of clusters
  
  numRows <- nrow(centroidMean)
  numCols <- ncol(centroidMean)
  
  newRows <- factorial(numRows)/(factorial(numRows - 2)*factorial(2))
  
  sqDistanceMatrix <- data.frame(newRows:numCols) 
  
  lister <- list()
  
  q = 1
  for (m in 2:nrow(centroidMean)-1){  ## loop for every row/cluster (except the last one)
    for (o in 1:nrow(centroidMean)) {  ## loop for list of row numbers
      lister[[o]] <- o
    }
    if (m == 1){ ## first time through the loop set this alternate list of rows according to the previous
      lister01 <- lister
    }
    lister[m] <- NULL  # Counts the total list minus the current row
    lister01[1] <- NULL  # Counts out rows previously used for comparison
    for(n in lister01){
      ComparatorRow <- n
      
      # lister02 <- unlist(lister01)
      # ComparatorRow <- lister02[1]
      # sqDistanceMatrix[q,1] <- paste0(m,":",ComparatorRow)   # annotate the new rows
      
      sqDistanceMatrix[q,1] <- paste0(m,":",ComparatorRow)   # annotate the new rows
      for (l in 2:ncol(centroidMean)){   # loop accross columns to perform formula for euclidian distance
        sqDistanceMatrix[q,l] <- (centroidMean[m,l] - centroidMean[ComparatorRow,l])^2  # first-row - comparator-row (squared)
      }
      q <- q + 1
    }
  }
    
  names(sqDistanceMatrix)[names(sqDistanceMatrix) == 'newRows.numCols'] <- 'clustPair'
  
  distances <- sqDistanceMatrix %>% 
    mutate(pairDistance = sqrt(rowSums(.[2:numCols])))
  
  
  Disters <- as.data.frame(distances[,numCols + 1])
  Widthers <- as.data.frame(clusterWidth[,numCols + 1])
  
  # ResolutionMatrix <- data.frame(45:45, stringsAsFactors = F) 
  ResolutionMatrix <- matrix(nrow = numCols, ncol = numCols)
  
  for (b in 1:nrow(Disters)){  ## loop for every pair
    pair <- distances[b,1]
    clustPair <- as_vector( strsplit(pair,split=":"))
    names(clustPair)[names(clustPair)==colnames(clustPair)] <- "currentPair"
    
    rownames(ResolutionMatrix[b]) <- clustPair[1]
    # names(ResolutionMatrix)[names(ResolutionMatrix)==rownames(ResolutionMatrix[b,])] <- clustPair[1]
    
    if (b <= numCols){  
      ResolutionMatrix[b,b] <- 0
    }
    ResolutionMatrix[as.numeric(clustPair[1]),as.numeric(clustPair[2])] <- Disters[b,1]/(Widthers[clustPair[1],] + Widthers[clustPair[2],])
    ResolutionMatrix[as.numeric(clustPair[2]),as.numeric(clustPair[1])] <- 0   # Set lower triangle to zeros
  }
  
  groupedInput <- tbl_df(as.data.frame(ResolutionMatrix))
  
  catList <- as.list(cats)
  orderCats <- sort(catList$Cell_Type)
  
  colnames( groupedInput ) <- orderCats
  
  suppressWarnings( rownames(groupedInput)  <- orderCats )
  
  groupedInput <- subset(groupedInput[1:numRows-1,2:numRows] )
  
  # scl <- "column"
  scl <- "none"
  hm <- pheatmap::pheatmap(groupedInput,
                           # hm <- pheatmap::pheatmap(groupedInput[,indexList],
                           # color = viridis,  # inferno(n=512, begin = 0, end = 0.9),
                           color = magma(n=512, begin = 0.05, end = 1.0),
                           border_color = "grey20",
                           main = "",
                           annotation_legend = F,
                           legend_labels = "Resolution Index",
                           show_rownames = TRUE,
                           show_colnames = TRUE,
                           fontsize_col = 12,
                           angle_col = "45",
                           fontsize_row = 12,
                           # labels_row = namerss,
                           scale = scl,
                           cluster_rows = F,
                           cluster_cols=F
                           )
  
  ggsave(paste0(outPath, "/ResolutionIndex_HeatMap.png"),  plot = hm, device = "png")
  
}


## TODO - Create a general function that computes resolution index:
resolution_index <- function(data, clusterDP) {
  
  
  
}
















    

