
## License ##
#######################################################################
# Copyright (c) 2020 Dr. Dr. Christoph Freier, & Ian Taylor BD Life 
# Sciences - Informatics, Ashland, OR
#
# R Script Template Designed and Written by: Josh Luthy and Ian Taylor
#
# License
# The software is distributed under the terms of the 
# Artistic License 2.0
# http://www.r-project.org/Licenses/Artistic-2.0
# 
# Disclaimer
# This software and documentation come with no warranties of any kind.
# This software is provided "as is" and any express or implied 
# warranties, including, but not limited to, the implied warranties of
# merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the  copyright holder be liable for any direct, 
# indirect, incidental, special, exemplary, or consequential damages
# (including but not limited to, procurement of substitute goods or 
# services; loss of use, data or profits; or business interruption)
# however caused and on any theory of liability, whether in contract,
# strict liability, or tort arising in any way out of the use of this 
# software.
######################################################################




#####
## Import
#####

outPath <- "FJ_CSV_OUPUT_FILE"
inCSVPath <- "FJ_DATA_FILE_PATH"
catParRaw <- "FJ_CAT_PAR"
newOutPut <- "FJ_DOWNLOAD_FILE_PATH"
timeStamp <- "FJ_TIMESTAMP"
runClustRCheckBool <- FJ_CLUST_R_CHECK
runTaylorIndexBool <- FJ_RESOLUTION_INDEX
exportFlag <- FJ_EXPORT_FLAG
maximum <- FJ_MAX

auto = F
if(maximum == 0){
    auto = T
}

#####
#####



####
## Functions
####

packages<-function(x){
    x<-as.character(match.call()[[2]])
    if(!require(x,character.only=TRUE)){
        install.packages(pkgs=x,repos="http://cran.r-project.org", type = 'binary')
        require(x,character.only=TRUE)
    }
}

devPackages<-function(x){
    if (!require(x,character.only=TRUE)){
        suppressMessages(suppressWarnings(devtools::install_github(x)))
    }else{
        require(x,character.only=TRUE)
    }
}

fCorePack<-function(x){
    x<-as.character(match.call()[[2]])
    if (!require(x,character.only=TRUE)){
        BiocManager::install(x)
        require(x,character.only=TRUE)
    }
}

stopAndReturnError <- function(message, ...) {
    cat(message, file=paste0(outs, "/Error_Message_from_R.txt"))
    stop(message, ...)
}

dip_it <- function(data, ...) {
    # if (!class(data) == "numeric")
    #   stop("data should be a numeric!")
    if (length(data) <= 1)
    stop("data length should be >1!")
    if (is.unsorted(data))
    data <- sort(data)
    n <- as.integer(length(data))
    result <- .C("diptst",
    x	      = as.double(data),
    n	      = n,
    dip      = double(1),
    lo.hi    = integer(4),
    ifault   = integer(1),
    gcm      = integer(n),
    lcm      = integer(n),
    mn	      = integer(n),
    mj	      = integer(n),
    debug    = as.integer(FALSE),
    min.is.0 = as.logical(FALSE))[3]
    return(result[[1]])
}

check_modality_it <- function(data,
cluster = "cluster", ...) {
    if(!is.data.frame(data))
    stop("Wrong input data, should be a data frame!")
    if(!cluster %in% colnames(data))
    stop("Missing input data, need to know cluster!")
    result <- data %>%
        rename(., cluster = cluster) %>%
        group_by(cluster) %>%
        summarise_if(is.numeric, list(~dip_it(.)))
    return(result)
}

heatmap_visual_it <- function(data, cluster = "cluster", ...) {
    if(!is.tbl(data))
    stop("Wrong input data, should be a data frame!")
    if(!"cluster" %in% colnames(data))
    stop("Missing input data, need to know cluster!")

    result00 <- data %>%
        rename(., cluster = cluster) %>%
        group_by(cluster) %>%
        rownames_to_column('clust')
    result <- result00 %>%
        group_by(clust) %>%
        arrange(order(as.factor(clust))) %>%
        select(-cluster) %>%
        rename(., cluster = clust) %>%
        pivot_longer(-cluster, names_to = "protein", values_to = "test") %>%
        ggplot(aes(x = protein, y = as.factor(as.integer(cluster)), fill = test)) +
        ggplot2::geom_tile(color = "white", size = .5) +
        scale_fill_gradient(low = "white", high = "red") + #, limits=c(0, (max(data))/20)) +
        theme_minimal() +
        coord_flip() +
        labs(x = "Parameters",
        y = cluster,
        title = paste0("Modality Score of ", cluster)) +
        theme(legend.title = element_blank())
    return(result)
}

heatmap_visual_it02 <- function(DATA, cluster = "cluster", ...) {
    # TODO - remove this:
    # DATA <- d
    # cluster <- catParam

    DATA <- DATA %>% arrange(-row_number())

    if(!is.tbl(DATA))
    stop("Wrong input data, should be a data tibble!")
    if(!"cluster" %in% colnames(DATA))
    stop("Missing input data, need to know cluster!")

    colNamer <- colnames(DATA)
    l <- ncol(DATA)
    n <- nrow(DATA)
    scen <- colNamer
    ls <- rep(1:n, length(scen)-1)
    dfe <- gather(DATA, key = "Parameters", value = "Modularity", 1:all_of(l))
    df <- subset(dfe, Parameters != "cluster")
    df$Clusters <- ls
    df$TModularity <- df$Modularity

    Parameters <- df$Parameters
    Parameter_Name <- colNamer[1:length(colNamer)-1]
    Clusters <- df$Clusters
    uniqueClusts <- unique(df$Clusters)
    # df

    # create the main plot, and take a look at it
    df.diff <- ggplot(df, aes(x = Parameters, y = Clusters, label = Parameters)) +
    # ggplot2::geom_tile(aes(fill = Modularity)) +
        ggplot2::geom_raster(aes(fill = Modularity), hjust = 0.5, vjust = 1.6) +
        scale_fill_viridis_c("Modularity", option = "plasma", end = 0.90)
    # df.diff

    h_total <- df %>%
        group_by(Clusters) %>%
        summarise(TModularity = sum(TModularity)) %>%
        mutate(Parameters = 'Total')

    v_total <- df %>%
        group_by(Parameters) %>%
        summarise(TModularity = sum(TModularity)) %>%
        mutate(Clusters = 'Total')

    h_mid <- max(h_total$TModularity)/2
    v_mid <- max(v_total$TModularity)/2

    if(h_mid > v_mid){
        mid = h_mid
    } else {
        mid = v_mid
    }

    p <- df.diff + new_scale_fill() + # ggplot(h_total, aes(label = TModularity)) +
        geom_tile(data = h_total, # Row sum
        aes(fill = TModularity ),
        size = 1,
        position = position_nudge(x = 0.7, y = 1.1)) +  # y = 1.6)) +
    # geom_text(data = h_total, size = 3, colour = "gray10", aes(label = round(TModularity,2))) + # -1.5) +  # Row sums
        geom_fit_text(data = h_total, size = 3, colour = "gray10", aes(label = round(TModularity,2))) +

        geom_tile(data = v_total, # ggplot(v_total, aes(label = TModularity)) +
        aes(fill = TModularity), # Col sum
        size = 1,
        position = position_nudge(x = -0.0, y = -0.4)) +
    # geom_text(data = v_total, size = 3, colour = "gray10", aes(label = round(TModularity,2)), angle = 45) + # Col sums
        geom_fit_text(data = v_total, size = 3, colour = "gray10", aes(label = round(TModularity,2)), angle = 45)+

        scale_fill_gradient2( # color palette option no. 1 (favored)
        "Modularity SUM",
        midpoint= mid,
        low="slateblue1",
        mid="violet",
        high="orange")
    # scale_color_gradient(low="blue", high="red") # Color palete option no. 2
    # scale_color_viridis_c(option = "plasma") # Color Pallete option no. 3

    q <- p +
        theme(panel.grid.major.x=element_blank(), #no gridlines
        panel.grid.minor.x=element_blank(),
        panel.grid.major.y=element_blank(),
        panel.grid.minor.y=element_blank(),
        panel.background=element_rect(fill="white"),
        axis.text.x = element_text(angle=45, hjust = -3,vjust= -3, size = 8,face = NULL),
        axis.text.y = element_text(hjust = -3,vjust=3, size = 8,face = NULL),
        plot.title = element_text(size=10,face="bold"),
        legend.title=element_text(face="bold", size=8))  +

        scale_x_discrete(name="Parameters", position = "top") +
        scale_y_discrete(name="Clusters", position = "left",
        limits = rev(c(unique(as.character(df$Clusters)), 'Total'))) +
    # labs(fill = "Modularity", color = "Modularity Total") +
        ggtitle(paste0("Modality Score of ", cluster)) +
        theme_minimal() +
        coord_fixed(expand = T) +
        theme(axis.text.x = element_text(angle=45, hjust = -0.1, vjust = -0.1))
    # q
    return (q)
}

arrange_col <- function(dl)
{
    n_col <- ncol(dl[[1]])
    out_names <- names(dl[[1]])
    out <- lapply(1:n_col, function(i, x) as.data.frame(lapply(x, '[[' , i = i)) , dl)
    names(out) <- out_names
    return(out)
}
#####
#####



#####
## Install
#####

packages(ellipsis)
packages(rlang)
packages(Rcpp)
packages(pkgload)
packages(backports)
packages(ps)
packages(tibble)
packages(devtools)
packages(tidyverse)
devPackages("Kawameicha/clustRcheck")

packages(clustRcheck)
packages(dplyr)
packages(tidyr)
packages(ggplot2)
packages(viridis)
packages(viridisLite)
packages(ggnewscale)
packages(ggfittext)
packages(colourvalues)
packages(pheatmap)

# packages(magma)
# packages(plasma)
# install.packages("devtools")
# devPackages("Kawameicha/clustRcheck")
# devtools::install_github("Kawameicha/clustRcheck")
# library(tidyverse)
# library(clustRcheck)

#####
#####


pdf(NULL)


#####
## Massage Imports
#####

noKatPar <- TRUE
if(catParRaw != "[]" && catParRaw != ""){
    catParam <- gsub("(\\[)","",gsub("(\\])","",catParRaw))
    catPar <- as.symbol(catParam)
    noKatPar <- FALSE
}

input00 <- read.csv(inCSVPath, sep = ",", header = TRUE, check.names = FALSE)

if(!nrow(input00) > 0){
    print("Unable to read the input file correctly, try again.")
    stopAndReturnError("Unable to read the input file correctly.", call. = FALSE)
}
if ("EventNumberDP" %in% colnames(input00)){
    print("EventNumber Detected - This is FlowJo")
    input00 <- input00 %>%
    select(EventNumberDP, everything())
    colnames(input00) <- gsub( "(^.*)( :: )" , "", colnames(input00), ignore.case = TRUE )
    isSeqGeqBool <- F
} else {
    print("We're in SeqGeq")
    isSeqGeqBool <- T
}

if (isSeqGeqBool == F) {
    input00$EventNumberDP <- NULL
}

## Check if the user selected any parameters (todo - unless no parameters is a valid choice in some case? - todo)
if (!length(input00) > 0) {
    stopAndReturnError("Forget something?...\nNo parameters for plotting detected in R.\n\nTry deleting and then
       restarting the plugin,\nmaking sure to select some parameters in\nthe top parameter selector dialog there.\n\nCheers!")
}
numUniqueParams <- length(unique(col(input00)))




#####
## Run
#####


# TaylorIndex
if(runTaylorIndexBool){   # Impliment functions for TaylorIndex

    # Lets take the selected Modal Parameter column and move it to the front of the DF before running gather()
    input0 <- input00 %>% dplyr::select(all_of(catParam), everything())

    # Get rid of pesky clusters at Zero (those should be cells not included in clustering which snuck in here)
    input <- subset(input0, input0[[catPar]]!=0)

    cats <- unique(input[catParam])

    colNamer <- colnames(input)

    m <- length(colNamer)

    centroids <- input %>%
        group_by_at(vars((!!catPar))) %>%
        summarise_all(mean)

    clusterWidth <- input %>%
        group_by_at(vars((!!catPar))) %>%
        summarise_all( sd ) %>%
        mutate(clusterWidth = sqrt(rowSums(.[2:m])^2))

    centroidMean <- as.data.frame(centroids)

    # n choose k: n!/[(n-k)!k!]
    # our k is always 2 (pairs)
    # this will be our new number of rows when comparing each possible pair of clusters

    numRows <- nrow(centroidMean)
    numCols <- ncol(centroidMean)

    newRows <- factorial(numRows)/(factorial(numRows - 2)*factorial(2))

    sqDistanceMatrix <- data.frame(newRows:numCols)

    lister <- list()

    q = 1
    for (m in 2:nrow(centroidMean)-1){  ## loop for every row/cluster (except the last one)
        for (o in 1:nrow(centroidMean)) {  ## loop for list of row numbers
            lister[[o]] <- o
        }
        if (m == 1){ ## first time through the loop set this alternate list of rows according to the previous
            lister01 <- lister
        }
        lister[m] <- NULL  # Counts the total list minus the current row
        lister01[1] <- NULL  # Counts out rows previously used for comparison
        for(n in lister01){
            ComparatorRow <- n

            sqDistanceMatrix[q,1] <- paste0(m,":",ComparatorRow)   # annotate the new rows
            for (l in 2:ncol(centroidMean)){   # loop accross columns to perform formula for euclidian distance
                sqDistanceMatrix[q,l] <- (centroidMean[m,l] - centroidMean[ComparatorRow,l])^2  # first-row - comparator-row (squared)
            }
            q <- q + 1
        }
    }

    sqDistanceMatrix <- na.omit(sqDistanceMatrix)

    names(sqDistanceMatrix)[names(sqDistanceMatrix) == 'newRows.numCols'] <- 'clustPair'

    distances <- sqDistanceMatrix %>%
    mutate(pairDistance = sqrt(rowSums(.[2:numCols])))


    Disters <- as.data.frame(distances[,numCols + 1])
    Widthers <- as.data.frame(clusterWidth[,numCols + 1])

    TaylorMatrix <- matrix(nrow = numRows, ncol = numRows)

    for (b in 1:nrow(Disters)){  ## loop for every pair
        pair <- distances[b,1]
        clustPair <- as_vector( strsplit(pair,split=":"))
        names(clustPair)[names(clustPair)==colnames(clustPair)] <- "currentPair"

        if (b <= numRows){
            TaylorMatrix[b,b] <- 0
        }
        TaylorMatrix[as.numeric(clustPair[1]),as.numeric(clustPair[2])] <- Disters[b,1]/(Widthers[clustPair[1],] + Widthers[clustPair[2],])
        TaylorMatrix[as.numeric(clustPair[2]),as.numeric(clustPair[1])] <- 0   # Set lower triangle to zeros
    }

    groupedInput <- tbl_df(as.data.frame(TaylorMatrix))

    colList <- list()
    for (i in 1:numRows) {
        colList[i] <- i
    }

    clustName <- unlist(colList)
    colnames(groupedInput) <- clustName
    suppressWarnings( rownames(groupedInput)  <- clustName )
    groupedInput <- subset(groupedInput[1:numRows-1,2:numRows] )



    # calculate the overall Taylor Score:
    Sum <- sum(TaylorMatrix)
    Clust <- n
    Cells <- nrow(input)
    Param <- ncol(input)
    ResIndex_Factor <- (sqrt(Cells) / (Clust*Param))*100
    ResScore <- log(Sum)*(ResIndex_Factor)^0.7

    if(!auto){
        breaklist = seq(0, maximum, by=0.01)
        # scl <- "column"
        scl <- "none"
        hm <- pheatmap::pheatmap(groupedInput,
        color = colourvalues::colour_values(breaklist, palette = "magma"),
        # color = colorRampPalette(brewer.pal(8, magma))(length(breaklist)),
        # color = magma(n=512, begin = 0.05, end = 1.0)(length(breaklist)),
        border_color = "grey20",
        main = paste0("TaylorIndex of ", catParam, " Pairs","\n","Overall TaylorScore: ", ResScore),
        annotation_legend = F,
        legend_labels = "TaylorIndex",
        show_rownames = TRUE,
        show_colnames = TRUE,
        fontsize_col = 12,
        angle_col = "45",
        fontsize_row = 12,
        # labels_row = namerss,
        scale = scl,
        cluster_rows = F,
        cluster_cols=F,
        breaks = breaklist
        )
    } else {
        # scl <- "column"
        scl <- "none"
        hm <- pheatmap::pheatmap(groupedInput,
        color = magma(n=512, begin = 0.05, end = 1.0),
        border_color = "grey20",
        main = paste0("TaylorIndex of ", catParam, " Pairs","\n","Overall TaylorScore: ", ResScore),
        annotation_legend = F,
        legend_labels = "TaylorIndex",
        show_rownames = TRUE,
        show_colnames = TRUE,
        fontsize_col = 12,
        angle_col = "45",
        fontsize_row = 12,
        # labels_row = namerss,
        scale = scl,
        cluster_rows = F,
        cluster_cols=F
        )
    }

    ## Export a figure
    ggsave(paste0(outPath, "/", catParam, "_TaylorIndex_HeatMap.png"),  plot = hm, device = "png")

    ## Export a table
    # if(exportFlag){
    #     if(isSeqGeqBool){
    #         write_csv(groupedInput,path = paste0(newOutPut, "/Euclid_",catParam,"_",timeStamp, "/TaylorIndex_",catParam, "_", timeStamp, ".csv"))
    #     } else {
    #         write_csv(d,path = paste0(outPath, "/TaylorIndex_",catParam, "_", timeStamp, ".csv"))
    #     }
    # }

}


## ClustRCheck
if(runClustRCheckBool){   # Impliment functions for ClustRCheck
    d <- check_modality_it(input00, cluster = catParam)
    p <- heatmap_visual_it02(d, cluster = catParam)
    ggsave(paste0(outPath, "/", catParam, "_ClustRCheck_HeatMap.png"),  plot = p, dpi = 320, device = "png")
    dev.off()

    ## Export a table
    if(exportFlag){
        if(isSeqGeqBool){
            write_csv(d,path = paste0(newOutPut, "/Euclid_",catParam,"_",timeStamp, "/ClustRCheck_",catParam, "_", timeStamp, ".csv"))
        } else {
            write_csv(d,path = paste0(outPath, "/ClustRCheck_",catParam, "_", timeStamp, ".csv"))
        }
    }
}




## Export a summary table
if(exportFlag){
    
    Sum <- sum(TaylorMatrix)
    
    Clust <- n
    Cells <- nrow(input)
    Param <- ncol(input)
    
    ResIndex_Factor <- (sqrt(Cells) / (Clust*Param))*100
    
    ResScore <- log(Sum)*(ResIndex_Factor)^0.7
    
    if(isSeqGeqBool){
        filler <- paste0(newOutPut, "/Euclid_",catParam,"_",timeStamp)
    } else {
        filler <- paste0(outPath)
    }
    
    resMat <- TaylorMatrix[1:(ncol(TaylorMatrix)-1),2:ncol(TaylorMatrix)]
    colnames(resMat) <- clustName[2:nrow(TaylorMatrix)]
    rownames(resMat) <- clustName[1:(ncol(TaylorMatrix)-1)]
    
    Cluster <- row.names(resMat)
    
    resMat <- as.data.frame(cbind(Cluster,resMat))
    
    write_csv(resMat,path = paste0(filler,"/TaylorIndex_perCluster_",catParam,"_",timeStamp,".csv"))
    
    cat(paste0("#Cells,", Cells, "\n", "#Parameters,", Param, "\n", "#Clusters,", Clust, "\n", "ScoreSum,", Sum,"\n",
               "ResolutionFactor,", ResIndex_Factor, "\n", "ResolutionScore,", ResScore), file=paste0(filler, "/Euclid_",
                                                                                                      catParam, "_SummaryScores", "_", timeStamp,".csv"))
}


## TODO - Create a general function that computes resolution index:
# resolution_index <- function(data, clusterDP) {
#
#
#
# }
